
from os.path import join
import time

from epics import caget, caput, PV

import iexcode.instruments.cfg as iex
from iexcode.instruments.utilities import dateandtime, print_warning_message
from iexcode.instruments.userCalcs import userStringSeq_pvs,userStringSeq_clear
from iexcode.instruments.shutters import shutter_check

def scan_fillin(VAL,RBV,start,stop,steps_points,**kwargs):
    """
    fills in the scan record for the curretn beamline ioc
    """
    try:
        iex.BL.mda.fillin(VAL,RBV,start,stop,steps_points,**kwargs)
    except:
        print("IEX_endstations has not been initialized")

def scan_fillin_table(VAL,RBV,my_table,**kwargs):
    """
    fills in the scan record for the curretn beamline ioc
    """
    try:
        iex.BL.mda.fillin_table(VAL,RBV,my_table,**kwargs)
    except:
        print("IEX_endstations has not been initialized")  

def scan_go(**kwargs):
    """
    Starts a scan 
    by default: scanDIM=1  
    Logging is automatic: use **kwargs or the optional logging arguments see scanlog() for details
    kwargs:
        X-ray = True (default), does shutter checks
             = False no shutter checks
    """
    try:
        iex.BL.mda.go(verbose=True,**kwargs)
    except:
        print("IEX_endstations has not been initialized")


def last_mda():
    """
    returns the last mda file number in the ioc defined by BL_ioc
    Previously: LastMDA
    """
    try:
        filenum = iex.BL.mda.lastFileNum()
        return filenum
    except:
        print("IEX_endstations has not been initialized")

def saveData_get_all(ioc_pv):
    """
    returns saveData info: 
    """
    try:
        saveData_info = {
            'fileSystem':caget(ioc_pv+"saveData_fileSystem",as_string=True),
            'subDir':caget(ioc_pv+"saveData_subDir",as_string=True),
            'scanNumber':caget(ioc_pv+"saveData_scanNumber"),
            'baseName':caget(ioc_pv+"saveData_baseName",as_string=True),
        }
        #take care of subDir starting with '/'
        f1 = saveData_info['fileSystem']  
        f2 = saveData_info['subDir'][1:] if saveData_info['subDir'][0] =='/' else saveData_info['subDir']
        filepath = join(f1,f2)
        #take care f filesystem starting with '//' (VME only)
        filepath=filepath.replace('//','/') 
        saveData_info['filepath'] = filepath
        return saveData_info
    except:
        print(ioc_pv+"not connected")
        return {}

   
    
##############################################################################################################
###########################                      Scan class                     ######################
##############################################################################################################
default_positioner_settling_time = 0.05
default_detector_settling_time = 0.1

class ScanRecord:
    """
    used for short hand scanning and to get info related to a scanRecord
    """

    def __init__(self,scan_ioc,**kwargs):
        """
        mda = scanRecord('29idTest:')
            mda.scan_ioc: full pv for the ioc, "29idARPES:", "29idc:"
            mda.detector_dictionary: dictionary corresponding to detector num and pv to proc => {det_num:'det_pv'}
            mda.trigger_dictionary: dictionary corresponding to trigger num and trigger pv => {trigger_num:'trigger_pv'}
            mda.before_scan_pv: pv to be proc'ed be for the start of a scan
            mda.after_scan_pv: pv to be proc'ed be for the start of a scan
            mda.log
        
        **kwargs:
        if a kwargs is not specified then it will not modify that field of the scanRecord

            empty = True
                kwargs = {
                    'detector_dictionary':'{}',
                    'trigger_dictionary':'{}',
                    'before_scan_pv':"",
                    'after_scan_pv':""
                }

            detector_dictionary
            trigger_dictionary
            before_scan_pv
            after_scan_pv
        """

        self.ioc = scan_ioc

        if 'empty' in kwargs:
            if kwargs['empty']:
                kwargs = {
                        'detector_dictionary':'{}',
                        'trigger_dictionary':'{}',
                        'before_scan_pv':"",
                        'after_scan_pv':""
                    }
        if 'detector_dictionary' in kwargs:
            self.detector_dictionary = kwargs['detector_dictionary']
        else:
            self.detector_dictionary = {}
        
        if 'trigger_dictionary' in kwargs:
            self.trigger_dictionary = kwargs['trigger_dictionary']
        else:
            self.trigger_dictionary = {}

        if 'before_scan_pv' in kwargs:
            self.before_scan_pv = kwargs['before_scan_pv']
        else:
            self.before_scan_pv = None
    
        if 'after_scan_pv' in kwargs:
            self.after_scan_pv = kwargs['after_scan_pv']
        else:
            self.after_scan_pv = None

        if 'default_detector_settling_time' in kwargs:
            self.default_detector_settling_time = kwargs['default_detector_settling_time']
        else:
            self.default_detector_settling_time = default_detector_settling_time

        if 'default_positioner_settling_time' in kwargs:
            self.default_positioner_settling_time = kwargs['default_positioner_settling_time']
        else:
            self.default_positioner_settling_time = default_positioner_settling_time
        
        if 'logging' in kwargs:
            self.log = kwargs['logging']
        else:
            self.log = None
        pass


        
        

##############################################################################################################
################################            MDA (ScanRecord) files and folders               ##############################
##############################################################################################################

    def fileNum(self): 
        """
        returns the current fileNum in the SaveData of ioc is scanning 
        or next if idle
        
        """
        
        return saveData_get_all(self.ioc)['scanNumber']

    def lastFileNum(self): 
        """
        returns the last saved fileNum in the SaveData of ioc

        previously: MDA_GetLastFileNum
        """

        fileNum = self.fileNum()
        return fileNum - 1


    def filepath(self):
        """
        returns the full filepath for the SaveData of ioc

        previously: MDA_CurrentDirectory
        """
        return saveData_get_all(self.ioc)['filepath']   

    def prefix(self):
        """
        returns the prefix for the SaveData of ioc

        previously: MDA_CurrentPrefix
        """
        return saveData_get_all(self.ioc)['baseName']   
   
    def current_run(self):
        """
        returns the current run by parsing the filepath in the SaveData for ioc

        previously: MDA_CurrentRun
        """
        filepath = self.filepath()
        m=filepath.find('data_29id')+len('data_29id')+2
        current_run=filepath[m:m+6]
        return current_run
    
    def user_name(self):
        """
        returns the current user by parsing the filepath in the SaveData for ioc

        previously: MDA_CurrentUser
        """
        subdir=saveData_get_all(self.ioc)['subDir']  
        m=subdir.find('/mda')
        if m == 0 : 
            current_user='staff'
        elif m > 0: 
            current_user=subdir[1:m]
        else: 
            current_user=""
            print_warning_message("mda_user is empty string")
        return current_user   


##############################################################################################################
#############################          filling in the scan record               ##############################
##############################################################################################################


    ### before and after scans
    def _before_after_scan_set(self,scan_dim,**kwargs):
        """
        Clears the before_scan and after_scan in all dims
        Fills in the before_scan and after_after pv for kwargs['scan_dim'] only
        **kwargs
            before_scan_pv
            after_scan_pv
        """

        scan_pv = self.ioc+"scan"+str(scan_dim)

        if 'before_scan_pv' in kwargs:
            self.before_scan_pv = kwargs['before_scan_pv']
            caput(scan_pv+".BSPV",self.before_scan_pv)
            caput(scan_pv+".BSCD",1)
            caput(scan_pv+".BSWAIT","Wait")

        if 'after_scan_pv' in kwargs:
            self.before_scan_pv = kwargs['after_scan_pv']
            caput(scan_pv+".ASPV",self.after_scan_pv)
            caput(scan_pv+".ASCD",1)
            caput(scan_pv+".ASWAIT","Wait")



    def default_after_scan_seq(self,**kwargs):
        """
        writes the user string sequence to happen at the end of a scan
        returns after_scan_pv = pv for userStringSeq for after scan

        uses links 1-4, link 10 is reserved for snake scans

        **kwargs
            seq_num: userStringSeq number in ioc => 10 (default)
            positioner_settling_time
            detector_settling_time

        Previously: AfterScan_StrSeq
        """
        kwargs.setdefault('seq_num',10)
        kwargs.setdefault('positioner_settling_time',default_positioner_settling_time)
        kwargs.setdefault('detector_settling_time',default_detector_settling_time)
        
        scan_dim = 1
        scan_pv = self.ioc+"scan"+str(scan_dim)

        seq_num=kwargs['seq_num']
        after_scan_pv = userStringSeq_pvs(self.ioc, seq_num)[0]
        self.after_scan_pv = after_scan_pv

        #clear and write the after scan user sequence
        userStringSeq_clear(self.ioc,seq_num)
        caput(after_scan_pv+".DESC","After Scan")

        #link 1 - clears positioners
        caput(after_scan_pv+".LNK1",scan_pv+'.P1AR'+" PP NMS")
        caput(after_scan_pv+".STR1","Clear pos&rdbk PV's, etc")

        #link 2 - sets positioner mode to to abosolute 
        caput(after_scan_pv+".LNK2",scan_pv+'.CMND'+" PP NMS")
        caput(after_scan_pv+".STR2","ABSOLUTE")

        #link 3 - sets positioner delay time to default 
        settling_time = kwargs['positioner_settling_time']
        caput(after_scan_pv+".LNK3",scan_pv+'.PDLY'+" PP NMS")
        caput(after_scan_pv+".DO3",settling_time)

        #link 4 - sets detector delay time to default 
        settling_time = kwargs['detector_settling_time']
        caput(after_scan_pv+".LNK4",scan_pv+'.DDLY'+" PP NMS")
        caput(after_scan_pv+".DO4",settling_time)

        return after_scan_pv+".PROC"
    
    ### Resetting:
    def reset_scan_dim(self,scan_dim,**kwargs):
        """
        Resets all the scanRecords (scanDIM=1,2,3,4) for a given IOC 
        **kwargs:
            detector_dictionary
            trigger_dictionary
            before_scan_pv
            after_scan_pv

        Previously: Reset_ScanAll
        """
           
        #clearing
        self.detectors_clear(scan_dim,{})
        self.triggers_clear(scan_dim,{})
        self.positioners_clear(scan_dim)

        #resetting
        if 'detector_dictionary' in kwargs:
            self.detectors_set(scan_dim,kwargs['detector_dictionary'])
        if 'trigger_dictionary' in kwargs:
            self.triggers_set(scan_dim,kwargs['trigger_dictionary'])

        self._before_after_scan_set(scan_dim,**kwargs)

        #settling time
        kwargs.setdefault('scan_dim',scan_dim)
        self.detector_settling_time(**kwargs)
        self.positioner_settling_time(**kwargs)

    def reset_all(self,**kwargs):
        """
        Resets all the scanRecords (scanDIM=1,2,3,4) for a given IOC 
        

        **kwargs
            detector_dictionary
            trigger_dictionary
            before_scan_pv
            after_scan_pv

        Note that for scan_dim = 1; default values are those that were set to attributes    

        Previously: Reset_ScanAll
        """
        kwargs.setdefault('detector_dictionary',self.detector_dictionary)
        kwargs.setdefault('trigger_dictionary',self.trigger_dictionary)
        if self.before_scan_pv != None:
            kwargs.setdefault('before_scan_pv',self.before_scan_pv)
        if self.after_scan_pv != None:
            kwargs.setdefault('after_scan_pv',self.after_scan_pv)
        
        #clearing all
        for scan_dim in range(4,1):
            self.reset_scan_dim(scan_dim,{},{},'','')

        #setting to defaults
        self.reset_scan_dim(1,**kwargs)

    ### default setting    
    def settings_defaults(self,scan_dim,verbose=True,**kwargs):
        """
        Reset scan settings to default: ABSOLUTE, STAY, 0.05/0.1 positioner/detector settling time
        
        Previous: Reset_Scan
        """
        kwargs.setdefault('scan_dim',scan_dim)
        scan_pv = self.ioc+"scan"+str(kwargs['scan_dim'])
        self.positioner_absolute_mode(absolute=True,**kwargs)      # Absolute position
        self.positioner_after_scan(after="PRIOR POS",**kwargs)   # After scan: prior position  
        self.positioner_settling_time(**kwargs)# Positioner Settling time => default
        self.detector_settling_time(**kwargs)# Detector Settling time => default
        self.triggers_reset(verbose=False,**kwargs)

        if verbose:
            print("Scan record "+self.ioc+" reset to default settings.")

    ### positioners 
    def positioners_clear(self,scan_dim,verbose=True):
        """
        Clear all extra scan positioners

        Previously: Clear_Scan_Positioners
        """
        scan_pv = self.ioc+"scan"+str(scan_dim)
        caput(scan_pv+".CMND",3)        # Clear all Positionners
        
        if verbose:
            print("\nAll extra positionners cleared")    

    def positioner_settling_time(self,**kwargs):
        """
        sets the positioner settling time
        """
        kwargs.setdefault('scan_dim',1)
        kwargs.setdefault('positioner_settling_time',self.default_positioner_settling_time)

        scan_pv = self.ioc+"scan"+str(kwargs['scan_dim'])

        caput(scan_pv+".PDLY",kwargs['positioner_settling_time'])     # Positioner Settling time
        return scan_pv+".PDLY",kwargs['positioner_settling_time']

    def positioner_after_scan(self,after="PRIOR POS",**kwargs):
        """
        sets the positioner after scan
        after = "PRIOR POS"
                "STAY"

        """
        kwargs.setdefault('scan_dim',1)
        scan_pv = self.ioc+"scan"+str(kwargs['scan_dim'])
        caput(scan_pv+".PASM",after)

    def positioner_absolute_mode(self,absolute=True,**kwargs):
        """
        absolute = True
                 = False => relative 

        """
        kwargs.setdefault('scan_dim',1)

        scan_pv = self.ioc+"scan"+str(kwargs['scan_dim'])
        if absolute:
            ab_state = 0
        else:
            ab_state = 1

        caput(scan_pv+".P1AR",ab_state)        # Absolute position

    def fillin(self,val_pv,rbv_pv,start,stop,steps_points,**kwargs):
        """
        Fills in the scanRecord for the positionner
            dim: scan dim
            val_pv: drive pv 
            rbv_pv: readback pv
            start: first point
            stop: last point
            step: step size (only used for positioner_num = 1, otherwise ignored)
        kwargs:
            num_points: False (default) => steps_points is the step size
                        True => steps_points is the number of points
            positioner_num: default => 1
            positioner_settling_time: default => 0.1
            detector_settling_time: default => 0.1

        Previously: Scan_FillIn
        """
        kwargs.setdefault('scan_dim',1)
        kwargs.setdefault("positioner_num",1)
        kwargs.setdefault('detector_settling_time',default_detector_settling_time)
        kwargs.setdefault('positioner_settling_time',default_positioner_settling_time)
        kwargs.setdefault('debug',False)

        self.progress(kwargs['scan_dim'])
    
        scan_pv = self.ioc+"scan"+str(kwargs['scan_dim'])
        posNum = kwargs['positioner_num']

        if kwargs['debug']:
            print('fillin: ',val_pv,rbv_pv,start,stop,steps_points)

        caput(scan_pv+".P"+str(posNum)+"PV",val_pv)
        caput(scan_pv+".R"+str(posNum)+"PV",rbv_pv)
        caput(scan_pv+".P"+str(posNum)+"SP",start*1.0)
        caput(scan_pv+".P"+str(posNum)+"EP",stop*1.0)
        
        if posNum == 1:
            if 'num_points' in kwargs:
                if kwargs['num_points'] :
                    caput(scan_pv+".NPTS",steps_points)
            else:
                caput(scan_pv+".P1SI",steps_points*1.0)
        
        self.detector_settling_time(**kwargs)
        self.positioner_settling_time(**kwargs)

        #checking that PVs and positioner limits are good
        self.check(kwargs['scan_dim'])


    ### detectors and detector settling time
    def detectors_clear(self,scan_dim,verbose=True):
        """
        Clear all scan detectors

        Previously: Clear_Scan_Detectors
        """
        scan_pv = self.ioc+"scan"+str(scan_dim)
        for i in range(1,10):
            caput(scan_pv+".D0"+str(i)+"PV","")
        for i in range(10,71):
            caput(scan_pv+".D"+str(i)+"PV","")
        if verbose:
            print("\nAll detectors cleared")

    def detectors_set(self,scan_dim,detector_dictionary):
        """
        sets the detectors
        detector_dictionary = {detNum:pv}
        """
        scan_pv = self.ioc+"scan"+str(scan_dim)
        for det_num in detector_dictionary.keys():
            det_pv=scan_pv+".D"+str(det_num).zfill(2)+"PV"
            caput(det_pv,detector_dictionary[det_num])

    def detector_settling_time(self,**kwargs):
        """
        set the detector settling time
        """
        kwargs.setdefault('detector_settling_time',self.default_detector_settling_time)

        scan_pv = self.ioc+"scan"+str(kwargs['scan_dim'])

        caput(scan_pv+".DDLY",kwargs['detector_settling_time'])
        return scan_pv+".DDLY",kwargs['detector_settling_time']

    ### triggers
    def triggers_set(self,scan_dim,trigger_dictionary):
        """
        Clear all scan detectors triggers
        
        trigger_dictionary = {trigger_num:pv}
        """
        scan_pv = self.ioc+"scan"+str(scan_dim)
        if len(trigger_dictionary.keys())>0:
            for tigger_num in  trigger_dictionary.keys():
                    caput(scan_pv+".T"+str(tigger_num)+"PV",trigger_dictionary[tigger_num])  

    def triggers_get(self,scan_dim):
        """
        returns a dictionary of the current scan striggers for scan_dim
        """
        scan_pv = self.ioc+"scan"+str(scan_dim)
        scan_trigger = {}
        for tigger_num in range(1,5):
            scan_trigger.update({tigger_num:caget(scan_pv+".T"+str(tigger_num)+"PV")})


    def triggers_clear(self,scan_dim,trigger_dictionary,verbose=True):
        """
        Clear all scan detectors triggers
        
        Previously: Clear_Scan_Triggers
        """
        scan_pv = self.ioc+"scan"+str(scan_dim)
        inner_scan_trigger = self.ioc+"scan"+str(scan_dim-1)+".EXSC"

        if len(trigger_dictionary.keys())>0:   
            for dim in range(1,5):
                if dim==1 and scan_dim>1:
                    caput(scan_pv+".T1PV",inner_scan_trigger)
                else:
                    caput(scan_pv+'.T'+str(dim)+'PV',"")    

    def triggers_reset(self,verbose=True,**kwargs):
        """
        sets the  trigger1 to be the lower dim execute, does not change the other triggers
        does nothing if scan_dim = 1
        """
        kwargs.setdefault('scan_dim',1)
        scan_pv = self.ioc+"scan"+str(kwargs['scan_dim'])

        inner_scan_trigger = self.ioc+"scan"+str(kwargs['scan_dim']-1)+".EXSC"
        if kwargs['scan_dim'] > 1:
            caput(scan_pv+".T1PV",inner_scan_trigger)  #Resets the higher dimensional scans to trigger the inner scan

        if verbose:
            print("Scan record "+self.ioc+"Trigger 1 is reset")


    ### checks
    def check(self,scan_dim):
        """
        checks pvs connects and that positioners are within limits

        returns scan_ok = True/False 
        """
        pv_ok = self._check_pvs(scan_dim)
        limits_ok = self._check_limits(scan_dim)
        
        scan_ok = bool(pv_ok * limits_ok)
        return scan_ok

    def _check_pvs(self,scan_dim,verbose=False):
        """
        Check if any of the detectors or positions pvs are not connected

        returns connected = True/False

        Previously: Scan_Check
        """
        if verbose:
            print('Checking if all detectors & positioners are connected...')

        scan_pv = self.ioc+"scan"+str(scan_dim)
        
        #Detectors
        det_ok = True
        for i in range(1,71):
            det_num=(str(i).zfill(2))
            det_pv=caget(scan_pv+".D"+det_num+"PV")
            if det_pv !='': 
                det=PV(det_pv); time.sleep(0.1)#smallest sleep to allow for PV traffic
                if not det.connected:
                    print("Detector "+det_num+" has a bad PV:  "+det.pvname+" not connected")
                    det_ok = False

        #Positioners
        pos_ok = True
        for i in range(1,5):
            pos_num=str(i)
            pos_pv=caget(scan_pv+".P"+pos_num+"PV")
            if pos_pv != '': 
                pos=PV(pos_pv); time.sleep(0.1)#smallest sleep to allow for PV traffic
                if not pos.connected:
                    print("Positioner "+pos_num+" has a BAD PV:  "+pos.pvname+" not connected")
                    pos_ok = False
        if verbose:
            print('Connected')
        
        return bool(pos_ok * det_ok)

    def _check_limits(self,scan_dim,verbose=False):
        """
        checks if the scan parameters are within the limits
        returns limits_ok (True / False)
        """
        scan_pv = self.ioc+"scan"+str(scan_dim)
        caput(scan_pv+'.CMND',1)
        time.sleep(.5)
        SMSG = caget(scan_pv+'.SMSG',as_string=True)
        if verbose:
            print(SMSG)

        if SMSG != 'SCAN Values within limits':
            print_warning_message('Positioner out of limits \n     '+SMSG)
            limits_ok = False
        else:
            limits_ok = True

        return limits_ok
    

    ##############################################################################################################
    #############################          table scans               ##############################
    ##############################################################################################################  
    def fillin_table(self,val_pv,rbv_pv,myarray,**kwargs):
        """
        Fills in the scan record for table scans given positioner=posNum
        myarray can be generated by myarray=Scan_MakeTable(StartStopStepLists)

        kwargs:
            positioner_num: default => 1
            scan_dim: default =>1
            detector_settling_time: 
            positioner_settling_time

        Previously: Scan_FillIn_Table
        """
        kwargs.setdefault("scan_dim",1)
        kwargs.setdefault("positioner_num",1)
        kwargs.setdefault('detector_settling_time',default_detector_settling_time)
        kwargs.setdefault('positioner_settling_time',default_positioner_settling_time)

        scan_dim = kwargs['scan_dim']

        self.progress(scan_dim)

        scan_pv = self.ioc+"scan"+str(scan_dim)
        posNum = kwargs['positioner_num']

        caput(scan_pv+".P"+str(posNum)+"SM","TABLE") 
        caput(scan_pv+".P"+str(posNum)+"PV",val_pv)         #Drive
        caput(scan_pv+".R"+str(posNum)+"PV",rbv_pv)         #Read  
        caput(scan_pv+".P"+str(posNum)+"PA",myarray)
        caput(scan_pv+'.NPTS',len(myarray))    

        self.detector_settling_time(**kwargs) 
        self.positioner_settling_time(**kwargs)
        
        #checking that PVs and positioner limits are good
        self.check(scan_dim)
  
    def table_reset_after(self,**kwargs):
        """
        resets positioner settling time 0.1
        sets all positionitonser to linear
        clears positioners

        Previously: Scan_Reset_AfterTable
        """
        #Setting everything back
        scan_pv = self.ioc+"scan"+str(kwargs['scan_dim'])

        self.detector_settling_time(**kwargs) 

        for i in range(1,5):
            caput(scan_pv+".P"+str(i)+"SM","LINEAR") 

        self.positioners_clear(self.ioc,kwargs['scan_dim'])

    ##############################################################################################################
    #############################         progress, go and abort             ##############################
    ##############################################################################################################
    def progress(self,scan_dim,verbose=True):
        """
        Checks if a scan is in progress, and sleeps until it is done
        
        pv now includes the full 29idARPES:
        Previously: Scan_Progress
        """
        pv = self.ioc+"scan"+str(scan_dim)
        check=caget(pv+".FAZE")
        while (check!=0):
            if verbose:
                print(pv+" in progress - please wait...")
            time.sleep(30)
            check=caget(pv+".FAZE")

    def abort(self):
        """
        aborts a scan for the command line

        Previously: Scan_Abort
        """
        caput(self.ioc+"AbortScans.PROC",1)

    def go(self,verbose=True,**kwargs):
        """
        checks limits and then starts a scan for a given ioc and scan_dim
        
        Previously: part of Scan_Go
        """
        kwargs.setdefault('scan_dim',1)
        scan_dim = kwargs['scan_dim']

        self.progress(scan_dim)

        if verbose:
            for i in range(1,scan_dim+1):
                drive = caget(self.ioc+"scan"+str(i)+".P1PV")
                start = caget(self.ioc+"scan"+str(i)+".P1SP")
                stop  = caget(self.ioc+"scan"+str(i)+".P1EP")
                step  = caget(self.ioc+"scan"+str(i)+".P1SI")
                print('Scan'+str(i)+': '+drive+'= '+str(start)+' / '+str(stop)+' / '+str(step))
        try: 
            if iex.BL.xrays:
                shutter_check(verbose=True)
        except:
            print_warning_message('no shutter check')
        if self.check(scan_dim):
            filename = self.prefix()
            fileNum  = self.fileNum()
            print(filename+str(fileNum)+" started at ", dateandtime())
            scan_pv = self.ioc+"scan"+str(scan_dim)
            caput(scan_pv+".EXSC",1,wait=True,timeout=900000)  #pushes scan button
            if self.log != None:
                self.log.update() # writes the log file
            print(filename+str(fileNum)+" finished at ", dateandtime())
            print('\n')
            

    ##############################################################################################################
    #############################          empty and time scans               ##############################
    ##############################################################################################################
 

    def empty_scan(self,npts=1,**kwargs):
        """
        starts a scan with out a drive
        npts = number of points

        **kwargs:
            execute = True/False (default=True)
            see scanRecord.fillin for more options

        Previously: Scan_Empty_Go
        """
        kwargs.setdefault('execute',True)
        kwargs.setdefault('num_points',True)
        self.fillin("","",0,npts-1,npts,**kwargs)   
        time.sleep(.1)

        if kwargs['execute']:
            self.go(**kwargs)
 
    def time_scan(self,duration_min,step_min=1,**kwargs):
        """
        starts a scan with the readback as time in seconds
        duration: total time for the scan in minutes
        step_min: steps per minute (sets the detector setting time for this)
        
        **kwargs
            execute True/False (default = True)

            trigger_dictionary = {1:'29idb:ca5:read'}
            
        
        Previously: Scan_Time_Go
        """
        kwargs.setdefault('execute',True)

        kwargs.update({'detector_settling_time':60.0/step_min})

        stop = duration_min 
        #num_pnts = duration_min * step_min
        dwell = 60/step_min
        self.fillin("","time",dwell/60,stop,dwell/60,**kwargs)

        print("Time scan - Settling time : "+str(step_min))
        if kwargs['execute']:
            self.go(**kwargs)

    ##############################################################################################################
    #############################    2D scans typewriter/snake scans               ##############################
    ##############################################################################################################
    def fillin_2D(self,inner_loop_list,outer_loop_list,**kwargs):
        """
        Fills in a the Scan Record for a 2D scan (Mesh),
        InnerMotorList=[val_pv1,rbv_pv1,start1,stop1,step1] #scanDIM=1
        OuterMotorList=[val_pv2,rbv_pv2,start2,stop2,step2] #scanDIM=2
        kwargs:
            num_points: False (default) => steps_points is the step size
                        True => steps_points is the number of points
            positioner_num: default => 1
            detector_settling_time: default => 0.1
            snake: coming soon            
        """
        kwargs.setdefault("snake",False)
        kwargs.setdefault("outer_scan_dim",2)

        outer_scan_dim = kwargs['outer_scan_dim']
        
        #fillin inner loop
        kwargs.update({'scan_dim':outer_scan_dim-1})
        self.fillin(inner_loop_list[0],inner_loop_list[1],inner_loop_list[2],inner_loop_list[3],inner_loop_list[4],**kwargs)
        #fillin outer loop
        kwargs.update({'scan_dim':outer_scan_dim})
        self.fillin(outer_loop_list[0],outer_loop_list[1],outer_loop_list[2],outer_loop_list[3],outer_loop_list[4],**kwargs)



    def snake_pv(self,snake_dim,enable=True):
        """
        used for snake scans (as opposed to typewriter scans)
        """
        snake_proc = self.snake_scan_userCalcOut(self.ioc,snake_dim=snake_dim,calc_num=1,enable=enable)
        return snake_proc


    def snake_scan_userCalcOut(self,snake_dim=1,calc_num=1,enable=False):
        """
        for snake scans you need to add this userCalcOut to the after scan
        it multiplies the step of the inner scan by -1 

        return the userCalcOut.PROC pv
        """
        snake_calc_pv = self.ioc+"userCalcOut"+str(calc_num) #Snake userCalc
        caput(self.ioc+"userCalcOutEnable.VAL",1) #global enable for userCalcs
        
        step_pv  = self.ioc+"scan"+str(snake_dim)+".P1SI"
        ### Setting up the user
        caput(snake_calc_pv+".DESC","Snake Calc")
        caput(snake_calc_pv+".INPA",step_pv+" NPP NMS")
        caput(snake_calc_pv+".CALC$","A*-1*B")
        caput(snake_calc_pv+".OUT",step_pv+" PP NMS")

        if enable:
            caput(snake_calc_pv+".B",1)
        else:
            caput(snake_calc_pv+".B",0)

        return snake_calc_pv+".PROC"