from numpy import inf,nan

from epics import caget, caput

import iexcode.instruments.cfg as iex
from iexcode.instruments.shutters import main_shutter_close
from iexcode.instruments.utilities import print_warning_message
from iexcode.instruments.encoders import encoders_reset_zero, _encoder_dictionary,encoder_sync 


slit_ioc="29idb:"

def _slits_dictionary():
    """
    dictionary of slit pv names for two and four blade slit assemblies
    key = slit_name
    val - slitHV
    """
    #'slit3C' : (None, "uses a flexure stage so is different"),
    
    d = {
        'slit1A' : ('29idb:Slit1H','29idb:Slit1V'),
        'slit2B' : ('29idb:Slit2H','29idb:Slit2V'),
        'slit3C' : ('29idb:Slit3C',),
        'slit3D' : ('29idd:Slit4V',),  
        }
    return d

def slit_name_list():
    d=_slits_dictionary()
    message = "slit_list = "
    for i, key in enumerate(d.keys()):
        if i==0:
            message += key
        else:
            message = message+", "+key
    print(message)

def _slits_wide_open_dictionary():
    """
    wide open size for the slits
    These are not the epics limits, but the physical extent of the beam
    """
    d = {
        'slit1A' : (4.5,4.5),
        'slit2B' : (6,8),
        'slit3C' : (300),
        'slit3D' : (2000),
        }
    return d

def slits_pvs(slit_name):
    """
    returns the rbv and drive for the size,center
    (rbv_size,val_size),(rvb_center,val_center)
    """

    d = _slits_dictionary()
    size = []#H,V
    center = []#H,V
    if slit_name in d.keys():
        if slit_name == 'slit3C':
            size_val = d[slit_name][0]+'Fit.A'
            size_rbv = d[slit_name][0]+'RBV'
            center_val = None 
            center_rbv = None
            size.append((size_val,size_rbv))
            center.append((center_val,center_rbv))
        else:
            for slit_HV in d[slit_name]:
                size_rbv = slit_HV +'size.VAL'
                size_val = slit_HV +'t2.C'
                center_rbv = slit_HV +'center.VAL'
                center_val = slit_HV +'t2.D'
                size.append((size_val,size_rbv))
                center.append((center_val,center_rbv))
    else:
        print_warning_message('slit_name = '+slit_name+' not a valid slit')
        slit_name_list()
    return size,center

def slits_synch(slit_name):
    """
    synch the motor position and the slit table for all the beamline slits
    """
    d = _slits_dictionary()
    if slit_name in _encoder_dictionary().keys():
        encoder_sync(slit_name)
    if slit_name in d.keys():
        if slit_name != 'slit3C':
            for slit_HV in d[slit_name]:
                slit_pv = slit_HV+'sync.PROC'
                caput(slit_pv,1)
    else:
        print_warning_message('slit_name = '+slit_name+' not a valid slit')
        slit_name_list()

def slits_sync_all():
    """
    synch the motor position and the slit table for all the beamline slits
    Slit-1A, slit-2B, slit-3C, slit-3D

    Previously: SyncAllSlits
    """
    slit_list = _slits_dictionary()
    for slit_name in slit_list.keys():
        slits_synch(slit_name)


def slits_get_all(verbose=True):
    """
    returns a dictionary with the current slit status
    """
    vals = {}
    for slit_name in ['slit1A','slit2B','slit3C','slit3D']:
        vals[slit_name+'_size'], vals[slit_name+'_center'] = slits_get(slit_name,verbose=verbose)
    return vals

def slits_set(slit_name,size,center,verbose=True,wait=True):
    """
    synchs the slit motors and the slit table and then sets the center and size
    
    slit_name is key in slits_dictionary
    size = (H_size, V_size)  #inf in direction specified => all the way open
    center = (H_center, V_center)
        None=> use current value
    
    Previously: SetSlit
    """
    d = _slits_dictionary()

    #syncing
    slits_synch(slit_name)

    if inf in size:
        size = _slits_wide_open_dictionary()[slit_name]
    
    size_pvs,center_pvs = slits_pvs(slit_name)
    
    for i in range(0,len(size_pvs)):
        #get pvnames
        size_rbv, size_val, = size_pvs[i]
        center_rbv,center_val = center_pvs[i]
        #check if there is a pv
        if center_rbv != None: 
            #check if specified value   
            if center[i] != None:
                caput(center_val, center[i],wait=wait,timeout=180)
        #check if there is a pv
        if size_rbv != None:  
            #check if specified value  
            if size[i] != None:
                caput(size_val, size[i],wait=wait,timeout=180)
    
    if verbose:
        slits_get(slit_name,verbose=True)

    return size,center


def slits_get(slit_name,verbose=True):
    """
    returns the current H_size,V_size,H_center,V_center
    
    """
    d=_slits_dictionary()
    size = []
    center = []

    #syncing
    slits_synch(slit_name)

    size_pvs,center_pvs = slits_pvs(slit_name)
    for i in range(0,len(size_pvs)):
        size_rbv, size_val, = size_pvs[i]
        center_rbv,center_val = center_pvs[i]
        if center_rbv != None:    
            center.append(caget(center_rbv))
        else:
            center.append(0)
        if size_rbv != None:  
            size.append(caget(size_rbv))



    if verbose:
        if len(d[slit_name])>1:
            print(slit_name + " = ("+str(round(size[0],3))+" x "+str(round(size[1],3))+") @ ("+str(round(center[0],3))+","+str(round(center[1],3))+")")
        else:
            print(slit_name + " = ("+str(round(size[0],3))+") @ ("+str(round(center[0],3))+")")

    return tuple(size),tuple(center)

def slits_set_size(slit_name,size):
    """
    sets the slit size with the current slit center

    size = (H,V) for slit1A, slit2B
    size = V for slit3D
    """
    size_current,center = slits_get(slit_name,verbose=False)
    slits_set(slit_name, size=size,center=center)

def slits_set_center(slit_name,center):
    """
    sets the slit size with the current slit center

    center = (H,V) for slit1A, slit2B
    center = V for slit3D
    """
    size,center_current = slits_get(slit_name,verbose=False)
    slits_set(slit_name,size=size,center=center)

def slits_scan_size(slit_name,direction,start,stop,step,**kwargs):
    """
    slit_name = 'slit1A','slit2B','slit3D'
    direction = 'H', 'V'
    **kwargs
        center: slits center,  if not specified then uses current center position
    """
    kwargs.setdefault('center',slits_get(slit_name,verbose=False)[1])
    kwargs.setdefault('execute',True) 

    size,center = slits_pvs(slit_name)  
    if direction == 'H':
        size_rbv,size_val = size[0]
        center_rbv,center_val = center[0]
    if direction == 'V':
        size_rbv,size_val = size[1]
        center_rbv,center_val = center[1]

    #set slit center
    slits_set_center(slit_name,kwargs['center'])
    #scan
    iex.BL.mda.fillin(size_rbv,size_val,start,stop,step,**kwargs)
    if kwargs['execute']:
        iex.BL.mda.go(**kwargs)

def slits_scan_center(slit_name,direction,start,stop,step,**kwargs):
    """
    slit_name = 'slit1A','slit2B','slit3D'
    direction = 'H', 'V'
    **kwargs
        size: slits size, if not specified then uses current size

    Previously: Scan_NarrowSlit_Go
    """
    kwargs.setdefault('size',slits_get(slit_name,verbose=False)[0])
    kwargs.setdefault('execute',True) 

    size,center = slits_pvs(slit_name) 
    if direction == 'H':
        size_rbv,size_val = size[0]
        center_rbv,center_val = center[0]
    if direction == 'V':
        size_rbv,size_val = size[1]
        center_rbv,center_val = center[1]
    
    #set slit center
    slits_set_size(slit_name,kwargs['size'])
    #scan
    iex.BL.mda.fillin(center_rbv,center_val,start,stop,step,**kwargs)
    if kwargs['execute']:
        iex.BL.mda.go(**kwargs)


def slit1A_scribe_marks():
    """
    moves slit1A to the position which corresponds to the scribe marks
    used to check if we have lost steps
    
    Dial values should be 0 on when on the scribe marks
    User values are set to correspond to the center of the beam

    Previously: GoToSlit1AScribe
    """
    main_shutter_close()
    for m in range(9,13):
        caput('29idb:m'+str(m)+'.DVAL',0,wait=True)

    print('At slit1A scribe marks') 
    
def slit1A_set(H_size,V_size,verbose=True,**kwargs):
    """
    sets the first aperture / slit to the specificed size
    if Hsize / Vsize = inf then goes to maximim 4.5 / 4.5
    **kwargs
        center = slit center, by default => (0,0)

    Previously: SetSlit1A
    """
    kwargs.setdefault('center',(None,None))
    slit_name = 'slit1A'

    size = (H_size,V_size)
    center = kwargs['center']
    slits_set(slit_name,size,center, verbose=verbose)
 
def slit1A_get(verbose=True):
    """
    returns the current (H_size,V_size),(H_center,V_center)
    """
    slit_name = 'slit1A'
    return  slits_get(slit_name,verbose=verbose)

def slit2B_set(H_size,V_size,verbose=True,**kwargs):
    """
    sets the clean up aperture / slit downstream of the mono to the specificed size
    if Hsize / Vsize = inf then goes to maximim 6 / 8
    
    **kwargs
        center = slit center, by default => (0,0)

    Previously: SetSlit2B
    """   
    kwargs.setdefault('center',(None,None))
    slit_name = "slit2B"

    size = (H_size,V_size)
    center = kwargs['center']
    slits_set(slit_name,size,center,verbose=verbose)

def slit2B_get(verbose=True):
    """
    returns the current (H_size,V_size),(H_center,V_center)
    """
    slit_name = "slit2B"   
    return  slits_get(slit_name,verbose=verbose)

def slit2B_encoders_reset(Hcenter,Vcenter):
    """
    Resetting Slit 2B encoders to 0 for a given position (Hcenter,Vcenter).
    Slit size need to be set to 0.

    Previously: Reset_Slit2B_Encoders
    """
    slit_name = 'slit2B'
    slits_set(slit_name,(0,0),(Hcenter,Vcenter), verbose=True)
    encoders_reset_zero(slit_name)
    slits_synch(slit_name)


def slit3C_set(size,verbose=True):
    """
    sets slit-3C (ARPES resolution defining slit)

    Previously SetSlit3C
    """
    slit_name = 'slit3C'
    center = None
    slits_set(slit_name,size,center, verbose=verbose)

def slit3C_get(size,verbose=True):
    """
    gets slit-3C (ARPES resolution defining slit)
    returns size, position
    """
    slit_name = "slit3C"   
    return  slits_get(slit_name,verbose=verbose)

def slit3D_set(V_size,position=None,verbose=True):
    """
    set the exit slit for the D-branch

    Previously: SetSlit3D
    """
    slit_name = "slit3D"

    if V_size in [inf,nan,None]: 
        V_size=2000

    size = (V_size)
    if position == None:
        center = slit3D_get()[1]
    else:
        center = (position)
        
    slits_set(slit_name,size,center,verbose=False)


    if position == None:
        size,center = slit3D_get()
        print("Slit-3D = "+str(size)+"um")
    else:
        print("Slit-3D = "+str(size)+'@ '+str(center)+"um")
        

def slit3D_get(verbose=True):
    """
    returns the current V_size,V_center
    """
    slit_name = "slit3D"
    (V_size),(V_center) = slits_get(slit_name,verbose)
    V_size = round(V_size,3)
    V_center = round(V_center,3)

    return  V_size,V_center

def slit3D_encoders_reset(Vcenter):
    """
    Resetting Slit 2B encoders to 0 for a given position (Hcenter,Vcenter).
    Slit size need to be set to 0.

    Previously: Reset_Slit2B_Encoders
    """

    slit_name = 'slit3D'
    slits_set(slit_name,(0),(Vcenter), verbose=True)
    encoders_reset_zero(slit_name)
    slits_synch(slit_name)

def exit_slit(size, verbose=True):
    """
    verbose used to supress printing

    Previously: SetExitSlit
    """
    branch=iex.BL.mda.branch
    if branch == "c":
        slit3C_set(size, verbose)
    elif branch == "d":
        slit3D_set(size, verbose)



##############################################################################################################
################################             slit fits             ##############################
##############################################################################################################

def aperture_fit(hv,slit_num):
    """
    used close the beamline apertures/slits to only take the center of the beam, 
        i.e. no heat bump (determined emperically by looking at the the shift in energy vs slit position)

    Previously Aperture_Fit
    """
    K=slit_coef(slit_num)[1]
    sizeH=K[0]+K[1]*hv+K[2]*hv*hv
    sizeV=K[3]+K[4]*hv+K[5]*hv*hv
    return [round(sizeH,3),round(sizeV,3)]

def slit_coef(slit_num):
    """
    3rd order polynomials coeffients for closing slit1A / slit2B to minimize heat bump effects
    slit_num = 1 / 2 for slit1A / slit2B
    """
    if slit_num == 1:
        pv='29id:k_slit1A'
        #Redshifted x (H):
        H0=2.3325
        H1=-.000936
        H2=2.4e-7
         #Redshifted z (V):
        V0=2.3935
        V1=-.0013442
        V2=3.18e-7
    if slit_num == 2:
        pv='29id:k_slit2B'
        #Redshifted x (H):
        H0=3.61
        H1=-0.00186
        H2=5.2e-7
        #Redshifted z (V):
        V0=6.8075
        V1=-0.003929
        V2=9.5e-7
    K=H0,H1,H2,V0,V1,V2
    return pv,K


def slit_3C_fit_calc_position(size):
    """
    used to convert slit size to motor position for slit 3C - flexure stage
    empirically determine offline with a camera, coeffients used by epics

    Previously: Slit3C_Fit
    """
    K0=-36.383
    K1=0.16473
    K2=-0.00070276
    K3=8.4346e-06
    K4=-5.6215e-08
    K5=1.8223e-10
    K6=-2.2635e-13
    motor=K0+K1*size+K2*size**2+K3*size**3+K4*size**4+K5*size**5+K6*size**6
    return motor

def slit_3C_fit_calc_size(position):
    """
    used to convert slit size to motor position for slit 3C - flexure stage
    empirically determine offline with a camera, coeffients used by epics
    
    Previously: Slit3C_RBV
    """
    K0=299.66
    K1=16.404
    K2=1.5572
    K3=0.14102
    K4=0.0064767
    K5=0.00014501
    K6=1.2617e-06
    size=round(K0+K1*position+K2*position**2+K3*position**3+K4*position**4+K5*position**5+K6*position**6,0)
    return size


